import ncs, _ncs
from _ncs.maapi import get_authorization_info, get_rollback_id
import re

DC_VNI_PREFIXES = {
    "mdc": 10000,
    "macc": 20000,
}

IOS_ABBRS = {
    "Fa": "FastEthernet",
    "Gi": "GigabitEthernet",
    "Te": "TenGigabitEthernet",
    "Twe": "TwentyFiveGigE",
    "Tw": "TwoGigabitEthernet",
    "Po": "Port_channel",
}

NXOS_ABBRS = {
    "Eth": "Ethernet",
    "Po": "port-channel",
}

EQDB_MODEL_TO_NED = {
    "^EX[234]200": "juniper-junos12-nc-4.7",
    "^EX3300": "juniper-junos12-nc-4.7",
    "^EX45[05]0": "juniper-junos12-nc-4.7",
    "^EX2300": "juniper-junos17-nc-4.7",
    "^EX3400": "juniper-junos17-nc-4.7",
    "^QFX": "juniper-junos17-nc-4.7",
    "^EX4600": "juniper-junos17-nc-4.7",
    "^WS-C": "cisco-ios-cli-6.80",
    "^CDB-8U$": "cisco-ios-cli-6.80",
    "^C9300": "cisco-ios-cli-6.80",
    "^ME-3800X": "cisco-ios-cli-6.80",
    "^C1000": "cisco-ios-cli-6.80",
    "^C3560": "cisco-ios-cli-6.80",
}

ISL_TRUNK_PLATFORMS = [
    "WS-C3560-24TS",
    "WS-C3560-48PS",
    "WS-C3560-48TS",
    "WS-C3560E-48TD",
    "WS-C3560-24PS",
    "WS-C3560G-24PS",
    "WS-C3560G-24TS",
    "WS-C3560G-48PS",
    "WS-C3560G-48TS",
    "WS-C3560G-48TS-S",
    "WS-C3560V2-48TS",
    "WS-C3560X-48",
    "WS-C3560X-48P",
    "WS-C3650-48PD",
    "WS-C3750-48P",
    "WS-C3750-48TS",
    "WS-C3750E-24PD",
    "WS-C3750E-48PD",
    "WS-C3750E-48TD",
    "WS-C3750G-12S",
    "WS-C3750G-24PS",
    "WS-C3750G-24TS",
    "WS-C3750G-24TS-1U",
    "WS-C3750G-48PS",
    "WS-C3750G-48TS",
    "WS-C3750X-48",
    "WS-C3750X-48P",
    "WS-C4503",
    "WS-C4506",
    "WS-C4506-E",
    "WS-C4507R",
    "WS-C4948-10GE",
    "WS-C4948E-F",
    "WS-C6504-E",
    "WS-C6506-E",
    "WS-C6509-E",
]

VPC_PORT_ASSIGNMENTS = {
    "C93360YC-FX2": ["Eth1/99", "Eth1/100"],
    "C93180YC-FX": ["Eth1/51", "Eth1/52"],
    "N3K-C36180YC-R": ["Eth1/51", "Eth1/52"],
    "C9504": ["Eth1/50", "Eth2/50"],
    "N3K-C3636C-R": ["Eth1/35", "Eth1/36"],
}

def model_to_ned_id(model):
    """
    Returns ned-id for a specific rancid platform
    """
    for pattern, ned_id in EQDB_MODEL_TO_NED.items():
        if re.match(pattern, model):
            return ned_id

    raise LookupError(f"No valid ned id for rancid device model {model}")


def get_users_groups(trans, uinfo):
    # Get the maapi socket
    s = trans.maapi.msock
    auth = get_authorization_info(s, uinfo.usid)
    return list(auth.groups)

def dns_to_device_name(dns_name: str) -> str:
    """Converts a DNS name to a device hostname"""

    return dns_name.replace(".umnet.umich.edu", "").lower()

def handle_commit(
    t, input, output, logging=None, commit_queue=False, no_networking=False
):
    """
    Generates the appropriate commit commands based on the umnet-action
    input - refer to dlzone-action-input grouping in actions.yang
    """
    if logging is not None:
        output.messages = logging.get_output_logs()

    if output.result == None:
        output.result = ""

    commit_params = ncs.maapi.CommitParams()

    # STEP 3: commit the transaction
    if input.dry_run.exists():
        if input.dry_run.outformat == "native":
            commit_params.dry_run_native()
            result = t.apply_params(True, params=commit_params)

            dry_run_output = result.get("device")
            if dry_run_output:
                # it's ugly - but would rather hand-build the string
                # then mess with the yang structure of the output to figure
                # out how to get the native data to look nice
                output.result += "device {\n"
                for device, cmds in dry_run_output.items():
                    output.result += f"\t{device}" + " {\n"
                    for cmd in cmds.split("\n"):
                        output.result += f"\t\t{cmd}\n"
                    output.result += "\t}\n"

                output.result += "}"

        elif input.dry_run.outformat == "xml":
            commit_params.dry_run_xml()
            result = t.apply_params(True, params=commit_params)
            output.result += str(result.get("local-node"))
        else:
            commit_params.dry_run_cli()
            result = t.apply_params(True, params=commit_params)
            output.result += str(result.get("local-node"))

    # for now we're doing "no networking" commits!
    elif ("no_networking" in input and input.no_networking) or no_networking:
        commit_params.no_networking()
        result = t.apply_params(True, params=commit_params)
        output.result = f"true: rollback ID {get_rollback_id(t.maapi.msock, t.th)}"

    elif commit_queue:
        commit_params.commit_queue_non_atomic()
        commit_params.commit_queue_async()
        result = t.apply_params(True, params=commit_params)
        output.result = f"true: commit queue ID: {result.get('id',False)}"

    else:
        result = t.apply()
        output.result = f"true: rollback ID {get_rollback_id(t.maapi.msock, t.th)}"

def hostname_to_codename(hostname: str) -> str:
    """
    generates a short codename from the provided hostname to be used in netnames

    'dl-arbl-1' -> 'ARBL1'
    """
    prefix, shortname, instance = hostname.split("-")
    return shortname.upper() + instance


def validate_existing_port(port: str, device: ncs.maagic.Node, log) -> bool:
    """
    Validates a port name against a device in the device tree.
    The inputted port should follow the abbreviation standards we use for cisco.
    The device should be a maagic node at /root/devices/device
    - look at /packages/umnet-ifconfig/python/umnet_ifconfig/validate.py

    """

    log.info(f"validating {device.name} {port}")
    ned_id = ncs.application.get_ned_id(device)

    # for junos all the ports are just keys into
    # /interfaces/interface
    if ned_id.startswith("juniper-junos"):
        port_list = device.config.configuration.interfaces.interface

    # for nxos and ios each interface type is its own container, and we
    # have to expand the interface name to match the right container name
    else:
        if ned_id.startswith("cisco-ios"):
            ABBRS = IOS_ABBRS
        elif ned_id.startswith("cisco-nx"):
            ABBRS = NXOS_ABBRS
        else:
            raise NotImplementedError(f"unsupported ned-id {ned_id}")

        for short, long in ABBRS.items():
            if port.startswith(short):
                port = port.replace(short, long)

        m = re.match("([A-Za-z_]+)(\d.*)$", port)
        if m:
            prefix = m.group(1)
            port = m.group(2)
            log.info(f"prefix and port: {prefix} {port}")
            port_list = getattr(device.config.interface, prefix)
        else:
            raise ValueError(f"Invalid port {port} for {device.name} (ned-id {ned_id})")

    if port in port_list:
        return True
    else:
        return False

def cli_write(trans: ncs.maapi.Transaction, msg: str):
    """
    output string to user CLI session.  trans must be the 'outtermost'
    transaction that the user opened when they invoked the action.
    """
    sock = trans.maapi.msock
    session_id = trans.maapi.get_my_user_session_id()
    return _ncs.maapi.cli_write(sock, session_id, msg + "\n")



import os
import logging
import multiprocessing
from typing import List, Dict
from functools import partial
from dataclasses import dataclass

import ncs
import multiprocessing_logging

from .models.junos import Junos, Junos12
from .models.nxos import NXOS
from .models.ios import IOS


def get_model(device: ncs.maagic.Node, log: logging.Logger) -> "BaseDevice":
    """
    returns the relevant state parsing model based on ned-id
    """
    ned_id = ncs.application.get_ned_id(device)
    root = ncs.maagic.get_root(device)
    if ned_id.startswith("cisco-ios"):
        return IOS(device, log)
    elif ned_id.startswith("juniper-junos12"):
        return Junos12(device, log)
    elif ned_id.startswith("juniper-junos17"):
        return Junos(device, log)
    elif ned_id.startswith("cisco-nx"):
        return NXOS(device, log)
    else:
        raise NotImplementedError(f"Unsupported ned-id {ned_id}")


def exec_cmds(dev_name: str, commands: List[str], results_q: multiprocessing.Queue):
    """
    open a new read transaction to query dev_name and save results to the
    multiproccessing queue.  Assumes that the calling application has verified
    /devices/device{dev_name} exists and is valid.

    :param dev_name: string name of an NCS device
    :param commands: list of model methods to call
    :param results: multiprocessing.Queue to save results to
    """
    pid = os.getpid()

    logger = logging.getLogger()
    ret = {"name": dev_name, "results": {}}

    with ncs.maapi.single_read_trans("admin", "system") as t:
        root = ncs.maagic.get_root(t)
        device = root.devices.device[dev_name]
        model = get_model(device, logger)  ## e.g. ../models/nxos.py

        # test and see if we're able to connect to the device before attempting
        # to issue any commands
        connection = device.connect()
        if not connection.result:
            logger.error(
                f"pid: {pid}: ERROR unable to connect to {dev_name}: {connection.info}"
            )
            err_str = f"Unable to connect to {dev_name}: {connection.info}"
            for cmd in commands:
                ret["results"].update({cmd: {"errors": err_str}})
            results_q.put(ret)
            return

        for cmd in commands:
            method = cmd.replace("-", "_")
            if hasattr(model, method):
                try:
                    ret["results"][cmd] = {}
                    invoke = getattr(model, method)
                    logger.debug(f"pid: {pid}: invoking {cmd} against {dev_name}")
                    data = invoke()
                    ret["results"][cmd]["data"] = data
                except Exception as e:
                    logger.error(
                        f"pid: {pid}: ERROR issuing {cmd} against {dev_name}: {e}"
                    )
                    ret["results"][cmd]["errors"] = str(e)
            else:
                ret["results"][cmd]["errors"] = f"{model}.{cmd} unsupported"

        # logger.debug(f"return results for {dev_name}: {ret}")
        results_q.put(ret)


class StateManager:
    """
    simple class to manage multiprocessing state when issuing commands to many devices.

    uses https://github.com/jruere/multiprocessing-logging to safely handle logging
    from child processes.

    USAGE:
        import umnet_pyncs

        with umnet_pyncs.state.StateManager() as m:
            interfaces = m.get_state(al_devices, ["get-interface-details"])
            arp = m.get_state(dl_devices, ["get-arp-table"])
            ...
    """

    def __init__(self):
        self.log = logging.getLogger()  ## root ncs pyvm logger object
        self.manager = multiprocessing.Manager()
        self.results_q = self.manager.Queue()

    def __enter__(self):
        # wrap root ncs logger instance with a queue
        multiprocessing_logging.install_mp_handler()
        return self

    def __exit__(self, *args, **kwargs):
        # unwrap the original ncs pyvm file handler now that we're done
        multiprocessing_logging.uninstall_mp_handler()

    def get_state(self, devices: List[ncs.maagic.Node], commands: List[str]) -> Dict:
        """
        returns results for all devices/commands as a dict keyed by device
        name, e.g.:

        { 'name': 'pe-ilab-1',
          'results': {
              'get-interface-details': [...],
              'get-bfd-neighbors':     [...],
              ...
          }
          'errors': None,
        },
        ...
        """

        ret = {}
        num_procs = len(devices) if len(devices) < os.cpu_count() else os.cpu_count()
        dev_names = [d.name for d in devices]

        with multiprocessing.Pool(processes=num_procs) as pool:
            self.log.info(f"submitting commands to worker pool ({num_procs})")
            pool.map(
                partial(exec_cmds, commands=commands, results_q=self.results_q),
                dev_names,
            )
            pool.close()

        # read results from the queue and structure return data for easier access
        while not self.results_q.empty():
            result = self.results_q.get()
            dev_name = result["name"]
            ret[dev_name] = {}
            ret[dev_name] = result["results"]

        return ret



from typing import List, Optional, Tuple
from ipaddress import IPv4Address

from ntc_templates.parse import parse_output
from netaddr import EUI, mac_unix_expanded

from .base import BaseDevice, ARPTableEntry, MACTableEntry, Interface, LLDPNeighbor
from ...utils import IOS_ABBRS

SPEED = {
    "Auto-speed": "auto",
    "10Mb/s": "10",
    "100Mb/s": "100",
    "1000Mb/s": "1000",
    "2.5Gb/s": "2.5G",
    "10Gb/s": "10G",
    "25Gb/s": "25G",
}

DUPLEX = {
"Half-duplex":"half",
"Full-duplex":"full",
}

def _remove_prompt(results: str) -> str:
    """
    remove the trailing prompt from the results of a live_status show
    command.  This only appears to be an issue for the IOS NED. i.e:

    ...
    Gi2/0/48        notconnect   1       auto   auto 10/100/1000BaseTX
    Twe2/1/1        connected    1       full    10G SFP-10GBase-SR
    Twe2/1/2        notconnect   1       auto   auto unknown
    Ap2/0/1         connected    1     a-full a-1000 App-hosting port
    al-shallow-1#    <-- (rude)
    """
    lines = results.splitlines()
    try:
        del lines[-1]
    except IndexError:
        pass
    return "\n".join(lines)


def _normalize_ifname(name: str) -> str:
    """
    Some IOS platforms return fully-qualified interface identifiers, e.g.
    'GigabitEthernet1/1' whereas others use abbreviated names e.g. 'Gi1\1'.
    Since the NCS switchport services uses abbreviated versions for IOS,
    this function will convert the names returned by the device to the
    shortened version
    """
    for abbr, full in IOS_ABBRS.items():

        if name.startswith(full):
            return abbr + name.split(full)[1]

    return name


class IOS(BaseDevice):
    def _run_cmd(self, command: str) -> str:
        """
        use NCS live-status exec to issue a raw show command towards a device.
        platform-specific models are expected to parse this output and return
        structured data.

        :param command: string CLI command, e.g. 'show interface status'
        :returns: raw string output from the device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = [command]

        results = show.request(inp)

        return _remove_prompt(results.result)

    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        """
        Gathers interface operational data from an IOS device by parsing 'show interfaces'

        :interface: optionally get data only for a single interface

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_interfaces

        ---
        parsed_sample:
        - interface: "FastEthernet1"
          link_status: "down"
          protocol_status: "down"
          hardware_type: "RP management port"
          address: "6c41.6aba.b47f"
          bia: "6c41.6aba.b47f"
          description: ""
          ...
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = f"interfaces {interface}" if interface else "interfaces"
        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show interfaces",
            data=reply,
        )

        ret = []
        for entry in parsed:
            input_errors = (
                0 if entry["input_errors"] == "" else int(entry["input_errors"])
            )
            output_errors = (
                0 if entry["output_errors"] == "" else int(entry["output_errors"])
            )
            record = Interface(
                name=_normalize_ifname(entry["interface"]),
                description=entry["description"],
                admin_status="disabled"
                if "administratively down" in entry["link_status"]
                else "enabled",
                oper_status="up" if "up" in entry["protocol_status"] else "down",
                duplex=DUPLEX.get(entry["duplex"], "auto"),
                speed=SPEED.get(entry["speed"], "auto"),
                input_errors=input_errors,
                output_errors=output_errors,
            )

            if record.name.startswith("Vlan"):
                record.is_logical = True

            ret.append(record)

        return ret

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        """
        Gathers ARP data from an IOS device by parsing the output of the CLI
        command 'show ip arp [ address | vrf <VRF>]'

        :address: optionally filter device output by MAC or IP address
        :vrf: optional string name of VRF table to look in

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_ip_arp

        TODO: add support(recursion?) for vrf='all' since ios makes you run a separate
              lookup command for every vrf (so rude)

        ---
        parsed_sample:
          - protocol: "Internet"
            address: "172.16.233.229"
            age: "-"
            mac: "0000.0c59.f892"
            type: "ARPA"
            interface: "Ethernet0/0"
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = "ip arp"
        if vrf is not None:
            command = command + f" vrf {vrf}"
        if address is not None:
            command = command + f" address {address}"
            normalized_address = IPv4Address(address)

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show ip arp",
            data=reply,
        )

        ret = []
        for entry in parsed:
            if entry["mac"] == "Incomplete":
                continue

            normalized_entry_address = IPv4Address(entry["address"])

            if address is None or normalized_address == normalized_entry_address:
                record = ARPTableEntry(
                    ip_address=IPv4Address(entry["address"]),
                    mac_address=EUI(entry["mac"], dialect=mac_unix_expanded),
                    interface=entry["interface"],
                    vrf=vrf,
                )
                ret.append(record)

        return ret

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        Gathers dynamically-learned MAC address data from an IOS device by
        parsing the output of the CLI command 'show mac address-table dynamic
        [address <address>]'

        :mac: optionally filter device output by MAC address

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_mac-address-table

        ---
        parsed_sample:
        - destination_address: "30a3.30a3.a1c3"
            type: "dynamic"
            vlan: "666"
            destination_port:
            - "Te1/30"
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = "mac address-table dynamic"
        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)
            command = command + f" address {mac}"

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show mac-address-table",
            data=reply,
        )

        ret = []
        for entry in parsed:
            try:
                vid = int(entry["vlan"])
            except ValueError:
                vid = None

            normalized_entry_mac = EUI(
                entry["destination_address"], dialect=mac_unix_expanded
            )
            # unless there's an on-going broadcast storm we should only be
            # learning an address on a single port -- not sure why the template
            # is modelled this way?
            interface = _normalize_ifname(entry["destination_port"][0])

            if mac is None or normalized_mac == normalized_entry_mac:
                record = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=interface,
                )
                ret.append(record)

        return ret

    def get_lldp_neighbors(self, interface: Optional[str] = None) -> List[LLDPNeighbor]:
        """
        Gathers active LLDP neighbors from an IOS device by
        parsing the output of the CLI command 'show lldp neighbors detail'

        :interface: optionally filter device output by interface

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/blob/master/tests/cisco_ios/show_lldp_neighbors_detail/cisco_ios_show_lldp_neighbors_detail1.yml

        ---
        parsed_sample:
        - local_interface: "Gi1/0/2"
          chassis_id: "7c25.86c9.aaaa"
          neighbor_port_id: "502"
          neighbor_interface: "ge-0/0/0.0"
          neighbor: ""
          system_description: "Juniper Networks, Inc. ex2200-24t-4g , version 12.3R9.4 Build\
          \ date: 2015-02-12 11:25:30 UTC"
          capabilities: "B,R"
          management_ip: ""
          vlan: "1"
          serial: ""
          power_pair: ""
          power_class: ""
          power_device_type: ""
          power_priority: ""
          power_source: ""
          power_requested: ""
        """
        # ncs automatically adds the 'show' to the front of the cmd
        command = "lldp neighbors detail"

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_ios",
            command="show lldp neighbors detail",
            data=reply,
        )

        ret = []
        for entry in parsed:
            record = LLDPNeighbor(
                local_interface=_normalize_ifname(entry["local_interface"]),
                remote_interface=_normalize_ifname(entry["neighbor_interface"]),
                remote_system_name=entry["neighbor"],
                remote_system_description=entry.get("system_description"),
            )
            if interface is None or entry["local_interface"] == interface:
                ret.append(record)

        return ret



from typing import List, Dict, Optional, Tuple
from ipaddress import IPv4Address

from ntc_templates.parse import parse_output
from netaddr import EUI, mac_unix_expanded

from .base import (
    BaseDevice,
    ARPTableEntry,
    MACTableEntry,
    Interface,
    LLDPNeighbor,
    EVPNMacIPEntry,
)
from ...utils import NXOS_ABBRS

SPEED = {
    "Auto-speed": "auto",
    "10 Mb/s": "10",
    "100 Mb/s": "100",
    "1000 Mb/s": "1000",
    "10 Gb/s": "10G",
    "25 Gb/s": "25G",
    "40 Gb/s": "40G",
    "100 Gb/s": "100G",
}
DUPLEX = {
    "half": "half",
    "full": "full",
}

def _normalize_ifname(name: str) -> str:
    """
    Some IOS platforms return fully-qualified interface identifiers, e.g.
    'GigabitEthernet1/1' whereas others use abbreviated names e.g. 'Gi1\1'.
    Since the NCS switchport services uses abbreviated versions for IOS,
    this function will convert the names returned by the device to the
    shortened version
    """
    for abbr, full in NXOS_ABBRS.items():

        if name.startswith(full):
            return abbr + name.split(full)[1]

    return name


class NXOS(BaseDevice):
    def _run_cmd(self, command: str) -> str:
        """
        use NCS live-status exec to issue a raw show command towards a device.
        platform-specific models are expected to parse this output and return
        structured data.

        :param command: string CLI command, minus the 'show', e.g. 'interface status'
        :returns: raw string output from the device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = [command]

        return show.request(inp)

    def get_interface_details(self, interface: Optional[str] = None) -> List[Interface]:
        """
        Gathers interface operational data from an NXOS device by parsing 'show interface'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface

        parsed_sample:
        - interface: "Ethernet1/1"
            link_status: "up"
            admin_state: ""
            hardware_type: "1000/10000 Ethernet"
            address: "8c60.4f4e.93a8"
            bia: "8c60.4f4e.93a8"
            description: "Uplink Fex131"
            ...
        """
        # ncs automatically adds the 'show' to the front of the cmd
        self.log.debug(f"sending 'show interface' to {self.device.name}")
        command = f"interface {interface}" if interface else "interface"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show interfaces",
            data=reply.result,
        )

        output = []
        for entry in parsed:
            input_errors = (
                0 if entry["input_errors"] == "" else int(entry["input_errors"])
            )
            output_errors = (
                0 if entry["output_errors"] == "" else int(entry["output_errors"])
            )

            i_name = _normalize_ifname(entry["interface"])
            interface = Interface(
                name=i_name,
                description=entry["description"],
                admin_status="enabled" if "up" in entry["admin_state"] else "disabled",
                oper_status="up" if "up" in entry["link_status"] else "down",
                duplex=DUPLEX.get(entry["duplex"], "auto"),
                speed=SPEED.get(entry["speed"], "auto"),
                input_errors=input_errors,
                output_errors=output_errors,
            )

            if interface.name.startswith("Vlan"):
                interface.is_logical = True

            output.append(interface)

        return output

    def get_bfd_neighbors(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers bfd operational data from an NXOS device by parsing 'show bfd neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://gitlab.umich.edu/grundler/ntc-templates/-/tree/add-nxos-show-bfd-neighbors/tests/cisco_nxos/show_bfd_neighbors

        ---
        parsed_sample:
          - our_addr: "172.23.0.18"
            neigh_addr: "172.23.0.19"
            ld_rd: "1090540255/1090566886"
            rh_rs: "Up"
            holddown: "862(3)"
            state: "Up"
            interface: "Eth1/10"
            vrf: "default"
            type: "SH"
        """
        self.log.debug(f"sending 'show bfd neighbors' to {self.device.name}")
        command = "bfd neighbors"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_ospf_neighbors(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers live ospf operational data from an NXOS device by parsing 'show
        ip ospf neighbor'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_ospf_neighbor

        ---
        parsed_sample:
          - interface: "Vlan999"
            local_ipaddr: "11.11.11.11"
            neighbor_ipaddr: "10.0.0.1"
            ospf_pid: "1111"
            state: "FULL/ -"
            uptime: "8w6d"
            vrf: "CUSTVRF1"
        """
        self.log.debug(f"sending 'show ip ospf neighbor' to {self.device.name}")
        command = "ip ospf neighbor"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_lldp_neighbors(self, interface: Optional[str] = None) -> List[LLDPNeighbor]:
        """
        Gathers live operational data about physically connected devices from an
        NXOS device by parsing 'show lldp neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_lldp_neighbors

        ---
        parsed_sample:
          - neighbor: "nx-osv9000-3-long-name.com"
            local_interface: "Eth1/1"
            neighbor_interface: "Ethernet1/1"
        """
        self.log.debug(f"sending 'show lldp neighbors' to {self.device.name}")
        command = "lldp neighbors detail"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show lldp neighbors detail",
            data=reply.result,
        )

        ret = []
        for entry in parsed:
            record = LLDPNeighbor(
                local_interface=entry["local_interface"],
                remote_interface=entry["neighbor_interface"],
                remote_system_name=entry["neighbor"],
                remote_system_description=entry.get("system_description"),
            )
            if interface is None or entry["local_interface"] == interface:
                ret.append(record)

        return ret

    def get_transceiver_details(self, interface: Optional[str] = None) -> List[Dict]:
        """
        Gathers live operational state DOM information from an NXOS device by
        parsing 'show interface tranceiver details'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface_transceiver_details

        ---
        parsed_sample:
          - interface: "Ethernet3/1"
            lane: ""
            temperature_value: "35.66"
            temperature_alarm_high: "75.00"
            temperature_alarm_low: "-5.00"
            temperature_warn_high: "70.00"
            temperature_warn_low: "0.00"
            voltage_value: "3.29"
            voltage_alarm_high: "3.63"
            voltage_alarm_low: "2.97"
            voltage_warn_high: "3.46"
            voltage_warn_low: "3.13"
            amps_value: "N/A"
            amps_alarm_high: "11.80"
            amps_alarm_low: "4.00"
            amps_warn_high: "10.80"
            amps_warn_low: "5.00"
            rx_value: "-30.45"
            rx_alarm_high: "2.00"
            rx_alarm_low: "-13.90"
            rx_warn_high: "-1.00"
            rx_warn_low: "-9.90"
            tx_value: "-14.05"
            tx_alarm_high: "1.69"
            tx_alarm_low: "-11.30"
            tx_warn_high: "-1.30"
            tx_warn_low: "-7.30"
        """
        self.log.debug(
            f"sending 'show interface transceiver details' to {self.device.name}"
        )
        command = "interface transceiver details"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_arp_table(
        self, address: Optional[str] = None, vrf: Optional[str] = None
    ) -> List[ARPTableEntry]:
        """
        Gathers live operational state ARP information from an NXOS device by
        parsing 'show ip arp detail vrf all'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_arp_detail

        NB: see https://github.com/networktocode/ntc-templates/pull/1204 -- this PR adds support for
        adding 'vrf all' to the CLI command which is required for our use case.

        ---
        parsed_sample:
          - address: "192.168.56.2"
            age: "00:17:02"
            mac: "5087.89a1.d8d5"
            interface: "Ethernet1/2"
            physical_interface: "Ethernet1/2"
            flags: "+"
            vrf: "PRIMARY"
          - address: "90.10.10.2"
            age: "00:02:55"
            mac: "000d.ece7.df7c"
            interface: "Vlan900"
            physical_interface: "Ethernet1/12"
            flags: ""
            vrf: "PRIMARY"
        """
        self.log.debug(f"sending 'show ip arp detail vrf all' to {self.device.name}")
        command = "ip arp detail vrf all"
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show ip arp detail vrf all",
            data=reply.result,
        )

        if address is not None:
            normalized_address = IPv4Address(address)

        output = []
        for entry in parsed:
            if entry["mac"] == "INCOMPLETE":
                continue

            if entry["physical_interface"] == "":
                interface = entry["interface"]
            else:
                interface = entry["physical_interface"]

            entry_vrf = None if entry["vrf"] == "" else entry["vrf"]
            normalized_entry_address = IPv4Address(entry["address"])

            if address is None or normalized_address == normalized_entry_address:
                arp = ARPTableEntry(
                    ip_address=IPv4Address(entry["address"]),
                    mac_address=EUI(entry["mac"], dialect=mac_unix_expanded),
                    interface=interface,
                    vrf=entry_vrf,
                )
                output.append(arp)

        return output

    def get_mac_table(self, mac: Optional[str] = None) -> List[MACTableEntry]:
        """
        Gathers learned MAC addresses from an NXOS device by parsing 'show mac address-table dynamic'.

        :mac: optionally filter device output by MAC address

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_mac_address-table

        ---
        parsed_sample:
          - vlan: "100"
            mac: "5087.89a1.de75"
            type: "static"
            age: "-"
            secure: "F"
            ntfy: "F"
            ports: "sup-eth1(R)"
        """
        command = "mac address-table dynamic"
        if mac is not None:
            normalized_mac = EUI(mac, dialect=mac_unix_expanded)
            command = command + f" address {mac}"

        self.log.debug(f"sending 'show {command}' to {self.device.name}")
        reply = self._run_cmd(command)
        # self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        parsed = parse_output(
            platform="cisco_nxos",
            command="show mac address-table",
            data=reply.result,
        )

        output = []
        for entry in parsed:
            try:
                vid = int(entry["vlan"])
            except ValueError:
                vid = None

            normalized_entry_mac = EUI(entry["mac"], dialect=mac_unix_expanded)

            if mac is None or normalized_mac == normalized_entry_mac:
                mac_entry = MACTableEntry(
                    address=normalized_entry_mac,
                    vlan_id=vid,
                    interface=entry["ports"],
                )
                output.append(mac_entry)

        return output

    def get_evpn_mac_ip(self, mac_or_ip: str = None) -> List[EVPNMacIPEntry]:
        """
        Parses "show bgp l2vpn evpn [route-type 2 | [mac_or_ip]]" and returns
        a list of evpn mac IP Entries.

        Note - this function doesn't validate the mac_or_ip input, we assume
        that it is well-formatted.
        """

        if mac_or_ip:
            command = f"bgp l2vpn evpn {mac_or_ip}"
        else:
            command = f"bgp l2vpn evpn route-type 2"

        reply = self._run_cmd(command)
        parsed = self._parse_textfsm(
            "cisco_nxos_show_bgp_l2vpn_evpn_macip", reply.result
        )

        # textfsm template returns a list of lists [mac, ip, vtep]
        # because of ecmp there will be duplicate entries.
        output = []
        for mac, ip, vtep in parsed:
            entry = EVPNMacIPEntry(
                mac_address=self.normalize_mac(mac), ip_address=ip, remote_vtep=vtep
            )
            if entry not in output:
                output.append(entry)

        return output
